/**********************************************************************
audio play & record (Linux, Solaris, IRIX)

$Id: playrec.c,v 1.7 1999/11/15 11:40:30 purnhage Exp $

Header file: playrec.h

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>
NM    Nikolaus Meine, Uni Hannover (c/o Heiko Purnhagen)
BE    Bernd Edler, Uni Hannover <edler@tnt.uni-hannover.de>

Changes:
16-sep-98   HP/NM   born, based on auplay.c and aurec.c
            HP/BE   ... and play_list.c
17-sep-98   HP      Solaris & IRIX
07-jan-99   HP      warning if no audio device support
08-jan-99   HP      audio device selection, fixed Solaris PROpenRead()
26-jan-99   HP      SGI: AL_DIGITAL_INPUT_RATE
18-feb-99   HP      write 0 before closing play device
23-feb-99   HP      fill up to n*numChan before closing play device
                    (prevents lock on Linux)
11-nov-99   HP      fixed debug output in PROpenWrite: device=
14-nov-99   HP      removed fileno(), using open() instead fopen()
**********************************************************************/

/* Audio play & record for Solaris, Linux, IRIX. */
/* Sample format: 16 bit twos complement, uniform quantisation */

/* Multi channel data is interleaved: l0 r0 l1 r1 ... */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "playrec.h"
#include "common.h"

#ifdef Linux
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#endif

#ifdef Solaris
#include <unistd.h>
#include <fcntl.h>
#include <sys/audioio.h>
#endif

#ifdef IRIX
#include <dmedia/audio.h>
#endif


/* ---------- declarations ---------- */

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define IRIXSIZE (1.0)		/* IRIX audio buffer size (sec) */


/* ---------- declarations (structures) ---------- */

struct PRStreamStruct		/* play/rec stream handle */
{
#ifdef Linux
  int fd;
#endif

#ifdef Solaris
  int fd;
#endif

#ifdef IRIX
  ALconfig rt_config;
  ALport rt_audio_p;
#endif

  long currentSample;		/* number of samples read/written */
  int write;			/* write (not read) flag */
  int numChannel;		/* number of channels */
  float fSample;		/* sampling frequency [Hz] */
};


/* ---------- variables ---------- */

static int PRdebugLevel = 0;	/* debug level */


/* ---------- functions ---------- */


/* PRInit() */
/* Init audio play & record. */

void PRInit (
  int debugLevel)		/* in: debug level */
				/*     0=off  1=basic  2=full */
{
  PRdebugLevel = debugLevel;
  if (PRdebugLevel >= 1) {
    printf("PRInit: debugLevel=%d\n",PRdebugLevel);
  }
}


/* PROpenRead() */
/* Open play/rec stream for reading. */

PRStream *PROpenRead (
  char *device,			/* in: device name (NULL = default device) */
  int *numChannel,		/* in/out: number of channels */
  float *fSample)		/* in/out: sampling frequency [Hz] */
				/*  numChannel & fSample: */
				/*  in: requested configuration */
				/*      (SGI: 0 = use audio panel settings) */
				/*  out: selected configuration */
				/*       (set according to audio */
				/*        device capabilities) */
				/* returns: */
				/*  play/rec stream (handle) */
				/*  or NULL if error */
{
  PRStream *s;

#ifdef Linux
  int err,tmp;
#endif  

#ifdef Solaris
  audio_info_t ainfo;
  int err;
#endif  

#ifdef IRIX
  long sr[2];
  long size;
  int err;
#endif  
  
  if (PRdebugLevel)
    printf("PROpenRead: device=\"%s\" numChannel=%d fSample=%.1f\n",
	   device?device:"(default)",*numChannel,*fSample);

  if ((s=(PRStream*)malloc(sizeof(PRStream)))==NULL)
    CommonExit(-1,"PROpenRead: Can not allocate memory");
  s->currentSample = 0;
  s->write = 0;

#ifdef Linux
  if (!device)
    device = "/dev/dsp";
  if ((s->fd=open(device,O_RDONLY,0))==-1) {
    CommonWarning("PROpenRead: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  tmp = AFMT_S16_LE;   /* aka 16 */
  err = ioctl(s->fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != AFMT_S16_LE) {
    CommonWarning("PROpenRead: 16 bit audio not supported");
    free(s);
    return NULL;
  }
  tmp = (*numChannel>1)?1:0;
  err = err || ioctl(s->fd, SNDCTL_DSP_STEREO, &tmp);
  *numChannel = tmp+1;
  tmp = (int)(*fSample+.5);
  err = err || ioctl (s->fd, SNDCTL_DSP_SPEED, &tmp);
  *fSample = tmp;
  if (err) {
    CommonWarning("PROpenRead: Unable to init dsp");
    free(s);
    return NULL;
  }
#endif

#ifdef Solaris
  if (!device)
    device = "/dev/audio";
  if ((s->fd=open(device,O_RDONLY,0))==-1) {
    CommonWarning("PROpenRead: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  AUDIO_INITINFO(&ainfo);

  ainfo.record.sample_rate = (long)(*fSample+.5);
  ainfo.record.channels = *numChannel;
  ainfo.record.precision	= 16;
  ainfo.record.encoding	= AUDIO_ENCODING_LINEAR;
  
  err = ioctl(s->fd,AUDIO_SETINFO,&ainfo);
  err = err || ioctl(s->fd,AUDIO_GETINFO,&ainfo);
  *fSample = ainfo.record.sample_rate;
  *numChannel = ainfo.record.channels;
  if (err) {
    CommonWarning("PROpenRead: Unable to init audio");
    free(s);
    return NULL;
  }
  if (ainfo.record.precision != 16 ||
      ainfo.record.encoding != AUDIO_ENCODING_LINEAR) {
    CommonWarning("PROpenRead: 16 bit audio not supported");
    free(s);
    return NULL;
  }
#endif

#ifdef IRIX
  sr[0] = AL_INPUT_RATE;
  sr[1] = (long)(*fSample+.5);
  err = 0;
  if (*fSample>0)
    err = err || ALsetparams(AL_DEFAULT_DEVICE,sr,2);
  err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
  *fSample = sr[1];
  if (*fSample == AL_RATE_AES_1) {
    sr[0] = AL_DIGITAL_INPUT_RATE;
    err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
    *fSample = sr[1];
  }
  s->rt_config = ALnewconfig();
  err = err || (!s->rt_config);
  err = err || ALsetsampfmt(s->rt_config,AL_SAMPFMT_TWOSCOMP);
  err = err || ALsetwidth(s->rt_config,AL_SAMPLE_16); 
  if (*numChannel>0)
    err = err || ALsetchannels(s->rt_config,*numChannel);
  err = err || ((*numChannel=ALgetchannels(s->rt_config))<0);
  size = (long)(*numChannel**fSample*IRIXSIZE);
  if (PRdebugLevel)
    printf("PROpenRead: got: numChannel=%d fSample=%.1f size=%ld\n",
	   *numChannel,*fSample,size);
  err = err || ALsetqueuesize(s->rt_config,size);
  if (err || !(s->rt_audio_p = ALopenport("adc","r",s->rt_config))) {
    CommonWarning("PROpenRead: Unable to init adc");
    free(s);
    return NULL;
  }
#endif

#if !(defined Linux || defined Solaris || defined IRIX)
  CommonWarning("PROpenRead: no audio device support");
  free(s);
  return NULL;
#endif

  s->numChannel = *numChannel;
  s->fSample = *fSample;

  if (PRdebugLevel)
    printf("PROpenRead: set to: numChannel=%d fSample=%.1f\n",
	   *numChannel,*fSample);

  return s;
}


/* PROpenWrite() */
/* Open play/rec stream for writing. */

PRStream *PROpenWrite (
  char *device,			/* in: device name (NULL = default device) */
  int *numChannel,		/* in/out: number of channels */
  float *fSample)		/* in/out: sampling frequency [Hz] */
				/*  numChannel & fSample: */
				/*  in: requested configuration */
				/*      (SGI: 0 = use audio panel settings) */
				/*  out: selected configuration */
				/*       (set according to audio */
				/*        device capabilities) */
				/* returns: */
				/*  play/rec stream (handle) */
				/*  or NULL if error */
{
  PRStream *s;

#ifdef Linux
  int err,tmp;
#endif  

#ifdef Solaris
  audio_info_t ainfo;
  int err;
#endif  

#ifdef IRIX
  long sr[2];
  long size;
  int err;
#endif  
  
  if (PRdebugLevel)
    printf("PROpenWrite: device=\"%s\" numChannel=%d fSample=%.1f\n",
	   device?device:"(default)",*numChannel,*fSample);

  if ((s=(PRStream*)malloc(sizeof(PRStream)))==NULL)
    CommonExit(-1,"PROpenWrite: Can not allocate memory");
  s->currentSample = 0;
  s->write = 1;

#ifdef Linux
  if (!device)
    device = "/dev/dsp";
  if ((s->fd=open(device,O_WRONLY,0))==-1) {
    CommonWarning("PROpenWrite: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  tmp = AFMT_S16_LE;   /* aka 16 */
  err = ioctl(s->fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != AFMT_S16_LE) {
    CommonWarning("PROpenWrite: 16 bit audio not supported");
    free(s);
    return NULL;
  }
  tmp = (*numChannel>1)?1:0;
  err = err || ioctl(s->fd, SNDCTL_DSP_STEREO, &tmp);
  *numChannel = tmp+1;
  tmp = (int)(*fSample+.5);
  err = err || ioctl (s->fd, SNDCTL_DSP_SPEED, &tmp);
  *fSample = tmp;
  if (err) {
    CommonWarning("PROpenWrite: Unable to init dsp");
    free(s);
    return NULL;
  }
#endif

#ifdef Solaris
  if (!device)
    device = "/dev/audio";
  if ((s->fd=open(device,O_WRONLY,0))==-1) {
    CommonWarning("PROpenWrite: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  AUDIO_INITINFO(&ainfo);

  ainfo.play.sample_rate = (long)(*fSample+.5);
  ainfo.play.channels = *numChannel;
  ainfo.play.precision	= 16;
  ainfo.play.encoding	= AUDIO_ENCODING_LINEAR;
  
  err = ioctl(s->fd,AUDIO_SETINFO,&ainfo);
  err = err || ioctl(s->fd,AUDIO_GETINFO,&ainfo);
  *fSample = ainfo.play.sample_rate;
  *numChannel = ainfo.play.channels;
  if (err) {
    CommonWarning("PROpenWrite: Unable to init audio");
    free(s);
    return NULL;
  }
  if (ainfo.play.precision != 16 ||
      ainfo.play.encoding != AUDIO_ENCODING_LINEAR) {
    CommonWarning("PROpenWrite: 16 bit audio not supported");
    free(s);
    return NULL;
  }
#endif

#ifdef IRIX
  sr[0]=AL_OUTPUT_RATE;
  sr[1]=(long)(*fSample+.5);
  err = 0;
  if (*fSample>0)
    err = err || ALsetparams(AL_DEFAULT_DEVICE,sr,2);
  err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
  *fSample = sr[1];
  s->rt_config = ALnewconfig();
  err = err || (!s->rt_config);
  err = err || ALsetsampfmt(s->rt_config,AL_SAMPFMT_TWOSCOMP);
  err = err || ALsetwidth(s->rt_config,AL_SAMPLE_16); 
  if (*numChannel>0)
    err = err || ALsetchannels(s->rt_config,*numChannel);
  err = err || ((*numChannel=ALgetchannels(s->rt_config))<0);
  size = (long)(*numChannel**fSample*IRIXSIZE);
  if (PRdebugLevel)
    printf("PROpenWrite: got: numChannel=%d fSample=%.1f size=%ld\n",
	   *numChannel,*fSample,size);
  err = err || ALsetqueuesize(s->rt_config,size);
  if (err || !(s->rt_audio_p = ALopenport("dac","w",s->rt_config))) {
    CommonWarning("PROpenWrite: Unable to init dac");
    free(s);
    return NULL;
  }
#endif

#if !(defined Linux || defined Solaris || defined IRIX)
  CommonWarning("PROpenWrite: no audio device support");
  free(s);
  return NULL;
#endif

  s->numChannel = *numChannel;
  s->fSample = *fSample;

  if (PRdebugLevel)
    printf("PROpenWrite: set to: numChannel=%d fSample=%.1f\n",
	   *numChannel,*fSample);

  return s;
}


/* PRReadData() */
/* Read data from play/rec stream. */

long PRReadData (
  PRStream *stream,		/* in: play/rec stream (handle) */
  short *data,			/* out: data[] */
  long numSample)		/* in: number of samples to be read */
				/* returns: */
				/*  number of samples read */
{
  long i;

  if (PRdebugLevel > 1)
    printf("PRReadData: numSample=%ld\n",numSample);

  if (stream->write)
    CommonExit(1,"PRReadData: device not in read mode");

#ifdef Linux
  i = read(stream->fd,data,sizeof(short)*numSample);
#endif

#ifdef Solaris
  i = read(stream->fd,data,sizeof(short)*numSample);
#endif

#ifdef IRIX
  i = (ALreadsamps(stream->rt_audio_p,data,numSample))?0:numSample;
#endif

  stream->currentSample += i;
  return i;
}


/* PRWriteData() */
/* Write data to play/rec stream. */

void PRWriteData (
  PRStream *stream,		/* in: play/rec stream (handle) */
  short *data,			/* in: data[] */
  long numSample)		/* in: number of samples to be written */
{
  if (PRdebugLevel > 1)
    printf("PRWriteData: numSample=%ld\n",numSample);

  if (!stream->write)
    CommonExit(1,"PRWriteData: device not in write mode");

#ifdef Linux
  write(stream->fd,data,sizeof(short)*numSample);
#endif

#ifdef Solaris
  write(stream->fd,data,sizeof(short)*numSample);
#endif

#ifdef IRIX
  ALwritesamps(stream->rt_audio_p,data,numSample);
#endif

  stream->currentSample += numSample;
}


/* PRClose() */
/* Close play/rec stream.*/

void PRClose (
  PRStream *stream)		/* in: play/rec stream (handle) */
{
  short data[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int n;

  if (PRdebugLevel)
    printf("PRClose: currentSample=%ld\n",stream->currentSample);

  n = stream->currentSample%stream->numChannel;
  if (n)
    CommonWarning("PRClose: odd number of samples (smps%%chan=%d chan=%d)",
		  n,stream->numChannel);

  if (stream->write) {
    if (n)
      n = stream->numChannel-n;		/* missing samples */
    n += stream->numChannel;		/* silence */
    n = min(16,n);			/* ugly, but ... */

#ifdef Linux
    write(stream->fd,data,sizeof(short)*n);
#endif

#ifdef Solaris
    write(stream->fd,data,sizeof(short)*n);
#endif

#ifdef IRIX
    ALwritesamps(stream->rt_audio_p,data,n);
#endif
  }

#ifdef Linux
  /*  ioctl(stream->fd, SNDCTL_DSP_SYNC, 0); */
  ioctl(stream->fd, SNDCTL_DSP_RESET, 0);
  close(stream->fd);
#endif

#ifdef Solaris
  close(stream->fd);
#endif

#ifdef IRIX
  if (stream->write)
    while (ALgetfilled(stream->rt_audio_p) > 0)
      sginap(1);
  ALcloseport(stream->rt_audio_p);
  ALfreeconfig(stream->rt_config);
#endif

  free(stream);
}


/* end of austream.c */

